// Based on http://wiki.apache.org/tapestry/Tapestry5HowToControlAccess

package jumpstart.web.services;

import java.io.IOException;

import jumpstart.business.domain.security.User;
import jumpstart.business.domain.security.iface.ISecurityFinderSvcLocal;
import jumpstart.web.annotation.ProtectedPage;
import jumpstart.web.state.theapp.Visit;

import org.apache.tapestry5.runtime.Component;
import org.apache.tapestry5.services.ApplicationStateManager;
import org.apache.tapestry5.services.ComponentClassResolver;
import org.apache.tapestry5.services.ComponentSource;
import org.apache.tapestry5.services.Dispatcher;
import org.apache.tapestry5.services.Request;
import org.apache.tapestry5.services.Response;
import org.slf4j.Logger;

/**
 * A service that protects pages annotated with {@link jumpstart.web.annotation.ProtectedPage}. It examines each
 * {@link org.apache.tapestry5.services.Request} and redirects it to the login page if the request is for a
 * ProtectedPage and the user is not logged in.
 * <p>
 * To use it, insert it as a service into Tapestry's dispatch chain as we do in AppModule.
 * 
 */
public class ProtectedPageGateKeeper implements Dispatcher {
	private final static String LOGIN_PAGE = "/theapp/login";
	private final String _autoLoginStr = System.getProperty("jumpstart.auto-login");

	private ApplicationStateManager applicationStateManager;
	private final ComponentClassResolver componentClassResolver;
	private final ComponentSource componentSource;
	private final IBusinessServicesLocator businessServicesLocator;
	private final Logger logger;

	/**
	 * Receive all the services needed as constructor arguments. When we bind this service, T5 IoC will provide all the
	 * services !
	 */
	public ProtectedPageGateKeeper(ApplicationStateManager asm, ComponentClassResolver resolver,
			ComponentSource componentSource, IBusinessServicesLocator businessServicesLocator, Logger logger) {
		this.applicationStateManager = asm;
		this.componentClassResolver = resolver;
		this.componentSource = componentSource;
		this.businessServicesLocator = businessServicesLocator;
		this.logger = logger;
	}

	public boolean dispatch(Request request, Response response) throws IOException {
		/*
		 * We need to get the Tapestry page requested by the user. So we parse the path extracted from the request
		 */
		String path = request.getPath();
		int nextslashx = path.length();
		String pageName;

		while (true) {
			pageName = path.substring(1, nextslashx);
			if (!pageName.endsWith("/") && componentClassResolver.isPageName(pageName))
				break;
			nextslashx = path.lastIndexOf('/', nextslashx - 1);
			if (nextslashx <= 1)
				return false;
		}
		return checkAccess(pageName, request, response);
	}

	/**
	 * Check the rights of the user for the page requested
	 * 
	 * @throws IOException
	 */
	public boolean checkAccess(String pageName, Request request, Response response) throws IOException {
		boolean canAccess = true;

		/* Is the requested page private ? */
		Component page = componentSource.getPage(pageName);
		boolean protectedPage = page.getClass().getAnnotation(ProtectedPage.class) != null;

		if (protectedPage) {
			canAccess = false;
			// If a Visit already exists in the session then you have already been authenticated
			if (applicationStateManager.exists(Visit.class)) {
				canAccess = true;
			}
			// Else if auto-login facility is on then auto-login as user 1 (secofr)
			// - this facility is for development environment only so you don't get thrown out of the
			// app every time you make a change that causes the session to be cleared as the server
			// does live page reloading.
			else if (isAutoLoginOn()) {
				autoLogin(1L);
				canAccess = true;
			}
		}

		/*
		 * This page can't be requested by a non-authenticated user => we redirect him to the LogIn page
		 */
		if (!canAccess) {
			response.sendRedirect(request.getContextPath() + LOGIN_PAGE);
			return true; // Make sure to leave the chain
		}

		return false;
	}

	/**
	 * Checks the value of system property jumpstart.auto-login. If "true" then returns true; if "false" then return
	 * false; if not set then returns false.
	 */
	private boolean isAutoLoginOn() {
		boolean autoLogin = false;
		if (_autoLoginStr == null) {
			autoLogin = false;
		}
		else if (_autoLoginStr.equalsIgnoreCase("true")) {
			autoLogin = true;
		}
		else if (_autoLoginStr.equalsIgnoreCase("false")) {
			autoLogin = false;
		}
		else {
			throw new IllegalStateException(
					"System property jumpstart.auto-login has been set to \""
							+ _autoLoginStr
							+ "\".  Please set it to \"true\" or \"false\".  If not specified at all then it will default to \"false\".");
		}
		return autoLogin;
	}

	/**
	 * Automatically logs you in as the given user. Its intention is to prevent you being thrown out of the application
	 */
	private void autoLogin(Long userId) {
		try {
			User user = getSecurityFinderService().findUser(userId);

			Visit visit = new Visit();
			visit.noteLogIn(user);
			logger.info(user.getLoginId() + " has been auto-logged-in.");

			applicationStateManager.set(Visit.class, visit);
		}
		catch (Exception e) {
			throw new IllegalStateException(e);
		}
	}

	private ISecurityFinderSvcLocal getSecurityFinderService() {
		return businessServicesLocator.getSecurityFinderSvcLocal();
	}
}
